
import unittest
from lastipy.lastfm.library import period
from lastipy.lastfm.library.scrobbled_track import ScrobbledTrack
from unittest.mock import patch, Mock
from requests import HTTPError
from lastipy.lastfm.library.top_tracks import fetch_top_tracks


class TopTracksFetcherTest(unittest.TestCase):
    
    @patch('requests.get')
    def test_one_page_of_results(self, mock_requests_get):
        expected_track = ScrobbledTrack(track_name="Stayin' Alive", artist="Bee Gees", playcount=2)

        mock_requests_get.ok = True

        json_track = self._build_json(expected_track)

        first_page_response = {
            'toptracks': {
                'track': [json_track]
            }
        }

        second_page_response = {
            'toptracks': {
                'track': []
            }
        }

        mock_responses = [Mock(), Mock()]
        mock_responses[0].json.return_value = first_page_response
        mock_responses[1].json.return_value = second_page_response
        mock_requests_get.side_effect = mock_responses

        self.assertEqual(fetch_top_tracks(api_key='', user='sonofjack3', a_period=period.SEVEN_DAYS)[0], expected_track)

    @patch('requests.get')
    def test_multiple_tracks_over_multiple_pages(self, mock_requests_get):
        expected_track_1 = ScrobbledTrack(track_name="Penny Lane", artist="The Beatles", playcount=5)
        expected_track_2 = ScrobbledTrack(track_name="Won't Get Fooled Again", artist="The Who", playcount=6)
        expected_track_3 = ScrobbledTrack(track_name="Like the FBI", artist="Bob Dylan", playcount=10)
        expected_tracks = [expected_track_1, expected_track_2, expected_track_3]

        mock_requests_get.ok = True

        json_track_1 = self._build_json(expected_track_1)
        json_track_2 = self._build_json(expected_track_2)
        json_track_3 = self._build_json(expected_track_3)

        first_page_response = {
            'toptracks': {
                'track': [json_track_1, json_track_2]
            }
        }

        second_page_response = {
            'toptracks': {
                'track': [json_track_3]
            }
        }

        third_page_response = {
            'toptracks': {
                'track': []
            }
        }

        mock_responses = [Mock(), Mock(), Mock()]
        mock_responses[0].json.return_value = first_page_response
        mock_responses[1].json.return_value = second_page_response
        mock_responses[2].json.return_value = third_page_response
        mock_requests_get.side_effect = mock_responses

        fetched_tracks = fetch_top_tracks(api_key='', user="sonofjack3", a_period=period.SEVEN_DAYS)
        self.assertCountEqual(fetched_tracks, expected_tracks)
    
    @patch('requests.get')
    def test_songs_with_one_playcount_ignored(self, mock_requests_get):
        ignored_track_1 = ScrobbledTrack(track_name="Stayin' Alive", artist="Bee Gees", playcount=1)
        non_ignored_track = ScrobbledTrack(track_name="Ventura Highway", artist="America", playcount=5)
        ignored_track_2 = ScrobbledTrack(track_name="Anesthetized Lesson", artist="Gum", playcount=1)

        mock_requests_get.ok = True

        json_track_1 = self._build_json(ignored_track_1)
        json_track_2 = self._build_json(non_ignored_track)
        json_track_3 = self._build_json(ignored_track_2)

        first_page_response = {
            'toptracks': {
                'track': [json_track_1, json_track_2, json_track_3]
            }
        }

        second_page_response = {
            'toptracks': {
                'track': []
            }
        }

        mock_responses = [Mock(), Mock()]
        mock_responses[0].json.return_value = first_page_response
        mock_responses[1].json.return_value = second_page_response
        mock_requests_get.side_effect = mock_responses

        fetched_tracks = fetch_top_tracks(api_key='', user='sonofjack3', a_period=period.SEVEN_DAYS)
        self.assertEqual(fetched_tracks.__len__(), 1)
        self.assertEqual(fetched_tracks[0], non_ignored_track)
    
    @patch('requests.get')
    def test_failure(self, mock_requests_get):
        mock_requests_get.ok = False
        mock_requests_get.side_effect = HTTPError("Mock")

        with self.assertRaises(HTTPError):
            fetch_top_tracks(api_key='', user='sonofjack3', a_period=period.SEVEN_DAYS)

    def _build_json(self, track):
        return {
            'name': track.track_name,
            'artist': {
                'name': track.artist
            },
            'playcount': track.playcount
        }



import logging
from lastipy.spotify.parse_spotify_tracks import parse_tracks

MAX_ITEMS_PER_REQUEST = 100


def get_tracks_in_playlists(spotify):
    """Returns all tracks in the currently-logged-in user's playlists"""

    logging.info("Fetching all tracks in " + spotify.current_user()['id'] + "'s playlists")

    playlists = spotify.current_user_playlists()['items']

    tracks = []
    for playlist in playlists:
        tracks = tracks + get_tracks_in_playlist(spotify, playlist_id=playlist['id'])

    logging.info("Fetched " + str(len(tracks)) + " tracks in " + str(len(playlists)) + " playlists")
    logging.debug("Fetched tracks: " + str(tracks))

    return tracks


def replace_tracks_in_playlist(spotify, playlist_name, tracks):
    """Replaces all tracks in the currently-logged-in user's given playlist 
        with the given tracks. If the playlist does not exist, creates it first."""

    logging.info("Replacing " + str(len(tracks)) + " tracks in " + spotify.current_user()['id'] + "'s playlist " + playlist_name)
    logging.debug("Replacing with tracks: " + str(tracks))

    playlist_id = _get_playlist(spotify, playlist_name)
    
    if playlist_id is None:
        playlist_id = _create_playlist(spotify, playlist_name)

    spotify.user_playlist_replace_tracks(user=spotify.current_user()['id'],
                                         playlist_id=playlist_id,
                                         tracks=[track.spotify_id for track in tracks])
                                        
    logging.info("Finished replacing tracks")


def get_tracks_in_playlist(spotify, playlist_name=None, playlist_id=None):
    """Returns all tracks in the given user's given playlist"""

    if playlist_id is None:
        if playlist_name is not None:
            logging.info("Fetching tracks in playlist with name " + playlist_name)
            playlist_id = _get_playlist(spotify, playlist_name)
        else:
            raise Exception("playlist_name or playlist_id required")
    else:
        logging.info("Fetching tracks in playlist with ID " + playlist_id)

    tracks_in_playlist = []
    keep_fetching = True
    while keep_fetching:
        json_response = spotify.playlist_tracks(playlist_id=playlist_id,
                                                offset=len(tracks_in_playlist))
        if json_response['items']:
            tracks_in_playlist = tracks_in_playlist + parse_tracks(json_response['items'])
        else:
            keep_fetching = False

    logging.info("Fetched " + str(len(tracks_in_playlist)) + " tracks")
    logging.debug("Fetched tracks: " + str(tracks_in_playlist))

    return tracks_in_playlist


def add_tracks_to_playlist(spotify, playlist_name, tracks):
    playlist_id = _get_playlist(spotify, playlist_name)

    if playlist_id is None:
        playlist_id = _create_playlist(spotify, playlist_name)
    
    logging.info("Adding " + str(len(tracks)) + " tracks to " + spotify.current_user()['id'] + "'s playlist " + playlist_name)
    logging.debug("Adding tracks: " + str(tracks))

    tracks_chunked = _chunk(tracks, MAX_ITEMS_PER_REQUEST)
    for chunk in tracks_chunked:
        # Remove any tracks that don't have a Spotify ID
        chunk = [track for track in chunk if track.spotify_id is not None]
        if len(chunk) > 0:
            spotify.user_playlist_add_tracks(user=spotify.current_user()['id'],
                                             playlist_id=playlist_id,
                                             tracks=[track.spotify_id for track in chunk])
        else:
            logging.debug("Chunk is empty, not adding any tracks")
    
    logging.info("Finished adding tracks")


def remove_tracks_from_playlist(spotify, playlist_name, tracks):
    logging.info("Removing " + str(len(tracks)) + " tracks from " + spotify.current_user()['id'] + "'s playlist " + playlist_name)
    logging.debug("Removing tracks: " + str(tracks))

    playlist_id = _get_playlist(spotify, playlist_name)

    if playlist_id is None:
        playlist_id = _create_playlist(spotify, playlist_name)

    
    tracks_chunked = _chunk(tracks, MAX_ITEMS_PER_REQUEST)
    for chunk in tracks_chunked:
        # Remove any tracks that don't have a Spotify ID
        chunk = [track for track in chunk if track.spotify_id is not None]
        if len(chunk) > 0:
            spotify.user_playlist_remove_all_occurrences_of_tracks(user=spotify.current_user()['id'],
                                                                   playlist_id=playlist_id,
                                                                   tracks=[track.spotify_id for track in chunk])
        else:
            logging.debug("Chunk is empty, not removing any tracks")
    
    logging.debug("Finished removing tracks")


def _chunk(lst, chunk_size):
    return [lst[i:i + chunk_size] for i in range(0, len(lst), chunk_size)] 


def _create_playlist(spotify, playlist_name):
    logging.info("Creating playlist " + playlist_name)
    return spotify.user_playlist_create(user=spotify.current_user()['id'], name=playlist_name)['id']


def _get_playlist(spotify, playlist_name):
    logging.debug("Looking for playlist " + playlist_name)
    playlists = spotify.current_user_playlists()
    playlists = [playlist for playlist in playlists['items'] if playlist['name'] == playlist_name]
    if playlists:
        return playlists[0]['id']
    else:
        return None



from lastipy.track import Track

class ScrobbledTrack(Track):
    """Represents a track that a user has scrobbled, with an associated playcount"""

    def __init__(self, track_name, artist, playcount):
        super().__init__(track_name, artist)
        self.playcount = playcount

    def __eq__(self, other):
        return isinstance(other, ScrobbledTrack) \
            and self.track_name == other.track_name \
            and self.artist == other.artist \
            and self.playcount == other.playcount

    def __hash__(self):
        return hash((self.track_name, self.artist, self.playcount))

    def __repr__(self):
        return str(self.__dict__)



from lastipy.spotify.parse_spotify_tracks import parse_tracks
from lastipy.spotify.playlist import get_tracks_in_playlist
import spotipy
import logging

MAX_ITEMS_PER_REQUEST = 50


def get_saved_tracks(spotify):
    """Returns the currently-logged-in users's saved tracks"""

    logging.info("Fetching " + spotify.current_user()['id'] + "'s saved tracks")

    saved_tracks = []
    keep_fetching = True
    while keep_fetching:
        json_response = spotify.current_user_saved_tracks(offset=len(saved_tracks))
        if json_response['items']:
            saved_tracks = saved_tracks + parse_tracks(json_response['items'])
        else:
            keep_fetching = False

    logging.info("Fetched " + str(len(saved_tracks)) + " saved tracks")
    logging.debug("Fetched tracks: " + str(saved_tracks))

    return saved_tracks


def add_tracks_to_library(spotify, tracks):
    logging.info("Adding " + str(len(tracks)) + " to " + spotify.current_user()['id'] + "'s library")
    logging.debug("Adding tracks: " + str(tracks))
    track_chunks = _chunk(tracks, MAX_ITEMS_PER_REQUEST) 
    for chunk in track_chunks:
        spotify.current_user_saved_tracks_add([track.spotify_id for track in chunk])
    logging.info("Finished adding tracks")


def remove_tracks_from_library(spotify, tracks):
    logging.info("Removing " + str(len(tracks)) + " from " + spotify.current_user()['id'] + "'s library")
    logging.debug("Removing tracks: " + str(tracks))
    track_chunks = _chunk(tracks, MAX_ITEMS_PER_REQUEST)
    for chunk in track_chunks:
        spotify.current_user_saved_tracks_delete([track.spotify_id for track in chunk])
    logging.info("Finished removing tracks")


def _chunk(lst, chunk_size):
    return [lst[i:i + chunk_size] for i in range(0, len(lst), chunk_size)] 